@@ -7,7 +7,7 @@ use FileHandle;
 
 use vars qw($CLASS $VERSION);
 $CLASS   = 'Config::Natural';
-$VERSION = '0.96_02';
+$VERSION = '0.99_02';
 
 my @base = (
     options => {
@@ -25,6 +25,8 @@ my @base = (
     state => {  }, 
     param => {  }, 
     handlers => {  }, 
+    prefilter => 0, 
+    filter => 0, 
 );
 
 # class option
@@ -45,11 +47,13 @@ for my $option (keys %{$base[1]}) {
 sub new {
     my $class = shift;
     my $self = bless { @base }, $class;
-    if(ref $_[0]) {
+    if(ref $_[0] eq 'HASH') {
         my $opts = shift;
         for my $option (keys %$opts) {
             $options{$option} = $opts->{$option} if exists $options{$option};
-            $self->{options}{$option} = $opts->{$option};
+            $self->{'options'}{$option} = $opts->{$option};
+            $self->filter($opts->{$option}) if $option eq 'filter';
+            $self->prefilter($opts->{$option}) if $option eq 'prefilter';
         }
     }
     $self->read_source(shift) if @_;
@@ -75,7 +79,7 @@ sub _get_set_option {
     my $option = shift;
     my $value  = shift;
     
-    carp "Unknown option '$option' " unless exists $self->{options}{$option} or $options{'quiet'};
+    carp "Unknown option '$option'" unless exists $self->{options}{$option} or $options{'quiet'};
     
     if(defined $value) {
         ($value, $self->{options}{$option}) = ($self->{options}{$option}, $value);
@@ -95,13 +99,13 @@ sub options {
     my @ret_list = ();
     
     for my $arg (@{$args->{'get'}}) {
-        carp "Class option '$arg' does not exist " and next 
+        carp "Class option '$arg' does not exist" and next 
             unless exists $options{$arg} or $options{'quiet'};
         push @ret_list, $options{$arg};
     }
     
     for my $arg (keys %{$args->{'set'}}) {
-        carp "Class option '$arg' does not exist " and next 
+        carp "Class option '$arg' does not exist" and next 
             unless exists $options{$arg} or $options{'quiet'};
         $options{$arg} = $args->{'set'}{$arg};
     }
@@ -169,9 +173,12 @@ sub read_source {
     my $state     = $self->{'state'};
     
     # store the name of the last opened file
-    $state->{filename} = $file;
+    $state->{'filename'} = $file;
     
     while(defined($_ = <$fh>)) {
+        ## execute the prefilter if present
+        $self->{'prefilter'} and $_ = &{$self->{'prefilter'}}($self, $_);
+        
         next if /^\s*$/;  # skip empty lines
         next if /^\s*$comment/;  # skip comments
         chomp;
@@ -179,7 +186,7 @@ sub read_source {
         ## include statement
         if(/^\s*\Q${include}\E\s+(\S+)\s*$/) {
             my $included = $1;
-            my @path = File::Spec->splitdir($state->{filename});
+            my @path = File::Spec->splitdir($state->{'filename'});
             pop @path;  # remove the current file name from the path
             $included = File::Spec->catdir(@path, $included);
             $self->read_source($included);
@@ -280,7 +287,7 @@ sub param {
     
     ## get the value of the desired parameters
     for my $arg (@{$args->{'get'}}) {
-        carp "Parameter '$arg' does not exist " and next
+        carp "Parameter '$arg' does not exist" and next
             if not exists $self->{'param'}{_case_($self, $arg)} and not $options{'quiet'};
         
         push @retlist, $self->{'param'}{_case_($self, $arg)}
@@ -292,10 +299,6 @@ sub param {
     my @arg_list = keys %{$args->{'set'}};
     
     if($current_list) {
-        unless(exists $self->{'param'}{$current_list}) {
-            $self->{'param'}{$current_list} = []
-        }
-        
         $param = ${$self->{'state'}{'lists_stacks'}}[-1];
         
     } else {
@@ -305,12 +308,15 @@ sub param {
     for my $arg (@arg_list) {
         my $value = $args->{'set'}{$arg};
         
-        ## check for a handler
-        if($self->has_handler($arg)) {
-            $param->{_case_($self, $arg)} = $self->exec_handler($arg, $value)
-        } else {
-            $param->{_case_($self, $arg)} = $value
-        }
+        ## use the filter if present
+        $self->{'filter'} and
+            $value = &{$self->{'filter'}}($self, $value);
+        
+        ## use the handler if present
+        $self->{'handlers'}{$arg} and 
+            $value = $self->exec_handler($arg, $value);
+        
+        $param->{_case_($self, $arg)} = $value
     }
     
     return wantarray ? @retlist : $retlist[0]
@@ -346,7 +352,7 @@ sub _parse_args {
                 }
                 
             } else {
-                carp "Bad ref $ref_type; ignoring it " unless $options{'quiet'};
+                carp "Bad ref $ref_type; ignoring it" unless $options{'quiet'};
                 next
             }
         
@@ -355,7 +361,7 @@ sub _parse_args {
            if(substr($arg, 0, 1) eq '-') {
                $arg = substr($arg, 1);
                my $val = shift;
-               carp "Undefined value for parameter '$arg' " and next 
+               carp "Undefined value for parameter '$arg'" and next 
                    if not defined $val and not $options{'quiet'};
                $args{'set'}{$arg} = $val if $arg
                
@@ -370,6 +376,32 @@ sub _parse_args {
 }
 
 
+# 
+# prefilter()
+# ---------
+# Set a new prefilter. 
+# 
+sub prefilter {
+    my $self = shift;
+    my $code = shift;
+    croak "Not a CODEREF" unless ref $code eq 'CODE';
+    $self->{'prefilter'} = $code;
+}
+
+
+# 
+# filter()
+# ------
+# Set a new filter. 
+# 
+sub filter {
+    my $self = shift;
+    my $code = shift;
+    croak "Not a CODEREF" unless ref $code eq 'CODE';
+    $self->{'filter'} = $code;
+}
+
+
 # 
 # set_handler()
 # -----------
@@ -378,8 +410,8 @@ sub _parse_args {
 sub set_handler {
     my $self = shift;
     my $param = shift;
-    my $sub = shift;
-    $self->{'handlers'}{$param} = $sub;
+    my $code = shift;
+    $self->{'handlers'}{$param} = $code;
 }
 
 
@@ -431,6 +463,38 @@ sub all_parameters {
 }
 
 
+# 
+# value_of()
+# --------
+# Return the value of the specified parameter
+# 
+sub value_of {
+    my $self = shift;
+    my $param_path = shift;
+    
+    # handle simple cases simply...
+    return $self->{'param'}{$param_path} if $self->{'param'}{$param_path};
+    
+    # handle more complex cases nicely.
+    my @path = split '/', $param_path;
+    not $path[0] and shift @path;
+    
+    my($name,$index) = ( (shift @path) =~ /^([^[]+)(?:\[([+-]?\d+)\])?$/ );
+    my $node = $self->param($name);
+    
+    if(ref $node) {
+        $node = $node->[int($index||0)];
+        for my $p (@path) {
+            ($name,$index) = ( ($p) =~ /^([^[]+)(?:\[([+-]?\d+|\*)\])?$/ );
+            $node = $node->{$name};  $index ||= 0;
+            ref $node and $index ne '*' and $node = $node->[int($index)];
+        }
+    }
+    
+    return $node
+}
+
+
 # 
 # delete()
 # ------
@@ -440,7 +504,7 @@ sub delete {
     my $self = shift;
     
     for my $param (@_) {
-        carp "Parameter '$param' does not exist " and next 
+        carp "Parameter '$param' does not exist" and next 
             if not exists $self->{'param'}{_case_($self, $param)} and not $options{'quiet'};
         delete $self->{'param'}{_case_($self, $param)}
     }
@@ -486,22 +550,60 @@ sub dump_param {
     my $args = _parse_args(@_);
     my $prefix = $args->{'set'}{'prefix'} || '';
     my $suffix = $args->{'set'}{'suffix'} || '';
-    my $nospace = $args->{'set'}{'nospace'} || '';
+    my $nospace = $args->{'set'}{'nospace'} || 0;
+
+    return _dump_tree($self, $self->{'param'}, 0, 
+        prefix => $prefix, suffix => $suffix, nospace => $nospace)
+}
+
+
+# 
+# _dump_tree()
+# ----------
+sub _dump_tree {
+    my $self = shift;
+    my $tree = shift;
+    my $level = shift;
+    my %state = @_;
     my $str = '';
     
-    for my $param (sort $self->all_parameters) {
-        next unless $param;
-        ## multi-line value ?
-        my $multiline = 1 if $self->param($param) =~ /\n|\r/;
+    if(ref $tree eq 'HASH') {
+        my $sp = $state{'nospace'} ? '' : ' ';
         
-        $str .= join '', $prefix, $param, 
-                ($nospace ? '' : ' '), 
-                $self->affectation_symbol, 
-                ($nospace ? '' : ' '), 
-                ($multiline ? $self->multiline_begin_symbol . $/ : ''), 
-                $self->param($param), 
-                ($multiline ? $self->multiline_end_symbol   . $/ : ''), 
-                $suffix, $/;
+        # add the list name and symbol
+        $state{'list_name'} and 
+        $str .= join '', 
+                $/, $sp x (($level-1)*2), $state{'list_name'}, $sp, 
+                $self->list_begin_symbol, $/;
+        
+        for my $param (sort keys %$tree) {
+            if(ref($tree->{$param})) {
+                $str .= _dump_tree($self, $tree->{$param}, $level+1, %state, list_name => $param)
+            
+            } else {
+                ## multi-line value?
+                my $multiline = 1 if $tree->{$param} =~ /\n|\r/;
+            
+                $str .= join '', 
+                        $sp x ($level*2), 
+                        $state{'prefix'}, $param, $sp, $self->affectation_symbol, $sp, 
+                        ($multiline ? $self->multiline_begin_symbol . $/ : ''), 
+                        $tree->{$param}, 
+                        ($multiline ? $self->multiline_end_symbol   . $/ : ''), 
+                        $state{'suffix'}, $/;
+            }
+        }
+       
+        # add the list end symbol
+        $state{'list_name'} and 
+        $str .= join '', 
+                $sp x (($level-1)*2), $self->list_end_symbol, $/;
+    
+    } elsif(ref $tree eq 'ARRAY') {
+        for my $list (@$tree) { $str .= _dump_tree($self, $list, $level, %state) }
+    
+    } else {
+        warn "unexpected reference type ", ref($tree)
     }
     
     return $str
@@ -528,6 +630,7 @@ sub write_source {
 1;
 
 __END__
+
 =head1 NAME
 
 Config::Natural - Module that can read easy-to-use configuration files
@@ -711,7 +814,7 @@ but instead
     }
 
 I don't think it's a big deal, because the aim of Config::Natural 
-is to be fast and read files with a clear and simple syntax. 
+is to be fast and to read files with a clear and simple syntax. 
 
 
 =head2 Inclusion
@@ -736,6 +839,99 @@ sharp sign C<"#">, it will be ignored. The sharp sign needs not
 being in the first column though. 
 
 
+=head1 SPECIAL FEATURES
+
+=head2 Filters
+
+Config::Natural offer three filter mechanisms so that you can modify 
+the data on-the-fly at two differents moments of the parsing. 
+
+   file.txt             read_source()
+   ___________          _________________
+  | ...       | =====> | reading file    |
+  | foo=hello |        | > for each line |     _____________
+  | ...       |        |        X <======|==> |  prefilter  |
+  |___________|        |        v        |    |_____________|
+                       |   parsing line -|--,
+                       |_________________|  |
+                                            |
+                        param()  <----------'
+                        _________________      _____________
+                       |         X <=====|==> | data filter |
+                       |         v       |    |_____________|
+                       |         X <=====|==> |   handler   |
+                       |         v       |    |_____________|
+                       |  storing value  |
+                       |_________________|
+
+=head2 Prefilter
+
+Prefilter occurs before Config::Natural parsing, therefore a prefilter 
+receives the current line as it was read from the file. This can be 
+used in order to correct some names which otherwise couldn't be parsed 
+by Config::Natural, for example names with spaces. Check in the 
+F<examples/> directory for sample programs that implements such functions. 
+
+You can set up a prefilter using the C<-E<gt>prefilter()> method, or at 
+creation time with the C<prefilter> option. 
+
+=head2 Data filter
+
+Data filter only occurs when affecting values to their parameters. 
+This can be used to implement additional features, like a syntax 
+for interpolating values. Check in the F<examples/> directory for 
+sample programs that implements such functions. 
+
+You can set up a data filter using the C<-E<gt>filter()> method, or at 
+creation time with the C<filter> option. 
+
+
+=head2 Handlers
+
+Handlers only occurs when affecting values to their parameters, 
+but instead of being object methods, handlers can be seen as 
+"parameters" methods, in that they are bound to a name, and are 
+only called when a parameter with that name is affected. 
+
+Handlers are defined with the C<-E<gt>handler()> method. 
+
+
+=head1 PARAMETER PATH
+
+This is a new functionality, introduced in version 0.99. 
+
+A parameter path is a way of referring any parameter, even if 
+it's deeply buried inside several layers of nested lists. 
+It is used by the method C<-E<gt>value_of()> to provide a much 
+easier way to read data hidden in nested lists. 
+
+The parameter path syntax is loosely inspired by XPath:
+
+    path = /level0[index0]/level1[index1]/.../param
+
+Indexes start at zero, like in Perl (and unlike XPath). 
+When an index is omitted, C<[0]> is assumed. 
+
+Examples: 
+
+    # same as $config->param('myparam')
+    $value = $config->value_of('/myparam');
+    
+    # same as $config->param('list')->[0]{myparam}
+    $value = $config->value_of(/list[0]/myparam);
+    $value = $config->value_of(/list/myparam);
+
+If you want to get back a whole list, instead of a single value, 
+use C<[*]> as the last index, and it will return the reference to 
+that list. 
+
+    # same as $config->param('list')
+    $value = $config->value_of('/list[*]');
+    
+    # same as $config->param('list')->[0]{inner_list}
+    $value = $config->value_of('/list/inner_list[*]');
+
+
 =head1 OBJECTS OPTIONS
 
 =head2 Syntax Options
@@ -812,8 +1008,23 @@ when reading a directory. Default is 0 (don't read hidden files).
 
 =item new ( )
 
-This method creates a new object. You can give an optional parameter, in 
-which case the read_source() method is called with that parameter. 
+=item new ( [ OPTIONS, ] FILE )
+
+This method creates a new object. 
+
+You can give an optional hashref in order to change settings of the 
+object at creation time. Any valid object option can be used here. 
+
+    my $config = new Config::Natural { read_hidden_files => 1 };
+
+You can also give a file name or a file handle, which will call 
+read_source() with that argument. 
+
+    # calling with a file name
+    my $config = new Config::Natural 'myconfig.conf';
+    
+    # calling with a file handle
+    my $config = new Config::Natural \*DATA;
 
 
 =item read_source ( I<FILENAME> )
@@ -897,6 +1108,13 @@ the value of the parameters of an object.
     );
 
 
+=item value_of ( PARAMETER_PATH )
+
+This method is an easier way to access the values of the parameters. 
+It returns the value of the parameter path given in argument. 
+Check L<"PARAMETER PATH"> for more information and some examples.
+
+
 =item all_parameters ( )
 
 This method returns the list of the parameters of an object.
@@ -963,6 +1181,32 @@ dump_param(). If no argument is given, the file or handle used by
 the last call of read_source() will be used. 
 
 
+=item filter ( I<CODEREF> )
+
+This method can be used to set a new data filter. 
+The subroutine code will be considered as an object method and 
+will receive the data as it was read. The return value of the 
+function will be used as the actual value. 
+For example: 
+
+    sub myfilter {
+        my $self = shift;  # important! remember it's a method
+        my $data = shift;
+        $data =~ s/\s*#.*$//go;  # remove comments appearing on 
+        return $data          # an affectation line
+    }
+    
+    my $conf = new Config::Natural { filter => \&myfilter };
+
+
+=item prefilter ( I<CODEREF> )
+
+This method can be used to set up a new input prefilter. 
+The same rules as for data filters applies, the only difference 
+being in that the prefilter can modify the data before 
+Config::Natural parses it. 
+
+
 =item set_handler ( I<PARAM, CODEREF> )
 
 This method can be used to hook a handler to a particular parameter. 
@@ -998,7 +1242,7 @@ SE<eacute>bastien Aperghis-Tramoni <sebastien@aperghis.net>
 
 =head1 COPYRIGHT
 
-Config::Natural is Copyright (C)2000-2003 SE<eacute>bastien Aperghis-Tramoni.
+Config::Natural is Copyright (C)2000-2004 SE<eacute>bastien Aperghis-Tramoni.
 
 This program is free software. You can redistribute it and/or modify it 
 under the same terms as Perl itself. 
